from decimal import Decimal
from math import isclose, isfinite
from pikepdf import _qpdf as qpdf, Pdf
from hypothesis import given, strategies as st, example
from hypothesis.strategies import (none, integers, binary, lists, floats, 
    characters, recursive, booleans)
import pytest


encode = qpdf._encode
decode = qpdf._decode


def decode_encode(obj):
    return decode(encode(obj))


def test_bool_involution():
    assert decode_encode(True) == True
    assert decode_encode(False) == False


@given(integers(min_value=-2**31, max_value=(2**31-1)))
def test_integer_involution(n):
    assert decode_encode(n) == n


@given(characters(min_codepoint=0x20, max_codepoint=0x7f))
@example('')
def test_ascii_involution(ascii):
    b = ascii.encode('ascii')
    assert decode_encode(b) == b


@given(binary(min_size=0, max_size=10000, average_size=10))
def test_binary_involution(binary):
    assert decode_encode(binary) == binary


@given(integers(-10**12, 10**12), integers(-10**12, 10**12))
def test_integer_comparison(a, b):
    equals = (a == b)
    encoded_equals = (qpdf._encode(a) == qpdf._encode(b))
    assert encoded_equals == equals

    lessthan = (a < b)
    encoded_lessthan = (qpdf._encode(a) < qpdf._encode(b))
    assert lessthan == encoded_lessthan


@given(integers(max_value=9223372036854775807), integers(max_value=9223372036854775807))
def test_crosstype_comparison(a, b):
    equals = (a == b)
    encoded_equals = (a == qpdf._encode(b))
    assert encoded_equals == equals

    lessthan = (a < b)
    encoded_lessthan = (qpdf._encode(a) < b)
    assert lessthan == encoded_lessthan


@given(integers(-10**12, 10**12), integers(0, 12))
def test_decimal_involution(num, radix):
    strnum = str(num)
    if radix > len(strnum):
        strnum = strnum[:radix] + '.' + strnum[radix:]

    d = Decimal(strnum)

    assert qpdf.Real(d).decode() == d


@given(floats())
def test_decimal_from_float(f):
    d = Decimal(f)
    if isfinite(f) and d.is_finite():
        py_d = qpdf.Real(d)
        assert isclose(py_d.decode(), d), (d, f.hex())
    else:
        with pytest.raises(ValueError, message=repr(f)):
            qpdf.Real(f)
        with pytest.raises(ValueError, message=repr(d)):
            qpdf.Real(d)


@given(lists(integers(-10, 10), min_size=0, average_size=5, max_size=10))
def test_list(array):
    assert decode_encode(array) == array


@given(lists(lists(integers(1,10), min_size=1, max_size=5),min_size=1,max_size=5))
def test_nested_list(array):
    assert decode_encode(array) == array


@given(recursive(none() | booleans(), lambda children: lists(children), max_leaves=20))
def test_nested_list(array):
    assert decode_encode(array) == array


def test_stack_depth():
    a = [150]
    for n in range(150):
        a = [a]
    assert decode_encode(a) == a
    with pytest.raises(RuntimeError, message="recursion"):
        encode(a) == encode(a)


import sys
import os
import platform

pytest_plugins = ['helpers_namespace']

import pytest
from pathlib import Path
from subprocess import Popen, PIPE


if sys.version_info < (3, 4):
    print("Requires Python 3.4+")
    sys.exit(1)


TESTS_ROOT = os.path.abspath(os.path.dirname(__file__))
PROJECT_ROOT = os.path.dirname(TESTS_ROOT)


@pytest.helpers.register
def check_refcount(thing, count):
    "Test that the reference count of thing is exactly 'count' in caller"
    try:
        from sys import refcount  # pypy doesn't haven't refcount
    except ImportError:
        return True  # ...so for pypy say it's okay

    # count + 1 because this function holds a reference, and its caller holds 
    # a reference, and we're writing this from the caller's perspective
    return refcount(thing) == count + 1


@pytest.fixture
def resources():
    return Path(TESTS_ROOT) / 'resources'


@pytest.fixture(scope="function")
def outdir(tmpdir):
    return Path(str(tmpdir))



import pytest
from pikepdf import _qpdf as qpdf

import os
import platform
import shutil
from contextlib import suppress


def test_create_form_xobjects(outdir):
    pdf = qpdf.Pdf.new()

    font = pdf.make_indirect(
        qpdf.Object.parse(b"""
            <<
                /Type /Font
                /Subtype /Type1
                /Name /F1
                /BaseFont /Helvetica
                /Encoding /WinAnsiEncoding
            >>"""))

    width, height = 100, 100
    image_data = b"\xff\x7f\x00" * (width * height)

    image = qpdf.Stream(pdf, image_data)
    image.stream_dict = qpdf.Object.parse("""
            <<
                /Type /XObject
                /Subtype /Image
                /ColorSpace /DeviceRGB
                /BitsPerComponent 8
                /Width 100
                /Height 100
            >>""")
    xobj_image = qpdf.Dictionary({'/Im1': image})

    form_xobj_res = qpdf.Dictionary({
        '/XObject': xobj_image
        })
    form_xobj = qpdf.Stream(pdf, b"""
        /Im1 Do
        """)
    form_xobj['/Type'] = qpdf.Name('/XObject')
    form_xobj['/Subtype'] = qpdf.Name('/Form')
    form_xobj['/FormType'] = 1
    form_xobj['/Matrix'] = [1, 0, 0, 1, 0, 0]
    form_xobj['/BBox'] = [0, 0, 1, 1]
    form_xobj['/Resources'] = form_xobj_res

    rfont = {'/F1': font}

    resources = {
        '/Font': rfont,
        '/XObject': {'/Form1': form_xobj},
        }

    mediabox = [0, 0, 612, 792]

    stream = b"""
        BT /F1 24 Tf 72 720 Td (Hi there) Tj ET
        q 144 0 0 144 234 324 cm /Form1 Do Q
        q 72 0 0 72 378 180 cm /Form1 Do Q
        """

    contents = qpdf.Stream(pdf, stream)

    page = pdf.make_indirect({
        '/Type': qpdf.Name('/Page'),
        '/MediaBox': mediabox,
        '/Contents': contents,
        '/Resources': resources
        })

    pdf.pages.append(page)
    pdf.save(outdir / 'formxobj.pdf')




import pytest
from pikepdf import Pdf, PdfError

import os
import platform
import shutil
from contextlib import suppress
from shutil import copy
import gzip
from pathlib import Path


# Files with unknown copyright status can't be shared publicly
PRIVATE_RESOURCES = Path(__file__).parent / 'resources' / 'private'


@pytest.fixture
def private():
    return Path(PRIVATE_RESOURCES)


pytestmark = pytest.mark.skipif(
    not PRIVATE_RESOURCES.is_dir(),
    reason='private resources not available')


def test_pypdf2_issue_361(private):
    with gzip.open(str(private / 'pypdf2_issue_361.pdf.gz'), 'rb') as gz:
        with pytest.raises(PdfError, match=r'trailer'):
            Pdf.open(gz)




import pytest
from pikepdf import _qpdf as qpdf, parse_content_stream, Page
import os
from subprocess import run, PIPE
import shutil


class PrintParser(qpdf.StreamParser):
    def __init__(self):
        super().__init__()

    def handle_object(self, obj):
        print(repr(obj))

    def handle_eof(self):
        print("--EOF--")


class ExceptionParser(qpdf.StreamParser):
    def __init__(self):
        super().__init__()

    def handle_object(self, obj):
        raise ValueError("I take exception to this")

    def handle_eof(self):
        print("--EOF--")


def test_open_pdf(resources):
    pdf = qpdf.Pdf.open(resources / 'graph.pdf')
    stream = pdf.pages[0]['/Contents']
    qpdf.Object.parse_stream(stream, PrintParser())


def test_parser_exception(resources):
    pdf = qpdf.Pdf.open(resources / 'graph.pdf')
    stream = pdf.pages[0]['/Contents']
    with pytest.raises(ValueError):
        qpdf.Object.parse_stream(stream, ExceptionParser())


@pytest.mark.skipif(
    shutil.which('pdftotext') is None, 
    reason="poppler not installed")
def test_text_filter(resources, outdir):
    input_pdf = resources / 'veraPDF test suite 6-2-10-t02-pass-a.pdf'

    # Ensure the test PDF has detect we can find
    proc = run(['pdftotext', str(input_pdf), '-'],
        check=True, stdout=PIPE, encoding='utf-8')
    assert proc.stdout.strip() != '', "Need input test file that contains text"

    pdf = qpdf.Pdf.open(input_pdf)
    stream = pdf.pages[0].Contents

    keep = []
    for operands, command in parse_content_stream(stream):
        if command == qpdf.Operator('Tj'):
            print("skipping Tj")
            continue
        keep.append((operands, command))

    new_stream = qpdf.Stream(pdf, keep)
    print(new_stream.read_bytes())
    pdf.pages[0]['/Contents'] = new_stream
    pdf.pages[0]['/Rotate'] = 90

    pdf.save(outdir / 'notext.pdf', True)

    proc = run(['pdftotext', str(outdir / 'notext.pdf'), '-'],
        check=True, stdout=PIPE, encoding='utf-8')

    assert proc.stdout.strip() == '', "Expected text to be removed"


def test_invalid_stream_object():
    with pytest.raises(TypeError):
        parse_content_stream(qpdf.Dictionary({"/Hi": 3}))


@pytest.mark.parametrize("test_file,expected", [
    ("fourpages.pdf", True),
    ("graph.pdf", False),
    ("veraPDF test suite 6-2-10-t02-pass-a.pdf", True),
    ("veraPDF test suite 6-2-3-3-t01-fail-c.pdf", False),
    ('sandwich.pdf', True)
])
def test_has_text(resources, test_file, expected):
    pdf = qpdf.Pdf.open(resources / test_file)
    for p in pdf.pages:
        page = Page(p)
        assert page.has_text() == expected



import pytest
from pikepdf import _qpdf as qpdf

import os
import platform
import shutil
import gc
from contextlib import suppress
from shutil import copy


def test_minimum_qpdf_version():
    assert qpdf.qpdf_version() >= '6.0.0'


def test_open_pdf(resources):
    pdf = qpdf.Pdf.open(resources / 'graph.pdf')
    assert '1.3' <= pdf.pdf_version <= '1.7'

    assert pdf.root['/Pages']['/Count'].as_int() == 1


def test_open_pdf_password(resources):
    pdf = qpdf.Pdf.open(resources / 'graph-encrypted.pdf', password='owner')
    assert pdf.root['/Pages']['/Count'].as_int() == 1


def test_open_pdf_wrong_password(resources):
    with pytest.raises(qpdf.PasswordError):
        qpdf.Pdf.open(resources / 'graph-encrypted.pdf', password='wrong')


def test_open_pdf_password_encoding(resources):
    with pytest.raises(qpdf.PasswordError):
        qpdf.Pdf.open(resources / 'graph-encrypted.pdf', password=b'\x01\xfe')


def test_open_pdf_no_password_but_needed(resources):
    with pytest.raises(qpdf.PasswordError):
        qpdf.Pdf.open(resources / 'graph-encrypted.pdf')


def test_stream(resources):
    with (resources / 'graph.pdf').open('rb') as stream:
        pdf = qpdf.Pdf.open(stream)
    assert pdf.root.Pages.Count == 1


def test_no_text_stream(resources):
    with pytest.raises(TypeError):
        with (resources / 'graph.pdf').open('r') as stream:
            qpdf.Pdf.open(stream)


def test_attr_access(resources):
    pdf = qpdf.Pdf.open(resources / 'graph.pdf')
    assert int(pdf.root.Pages.Count) == 1


def test_create_pdf(outdir):
    pdf = qpdf.Pdf.new()

    font = pdf.make_indirect(
        qpdf.Object.parse(b"""
            <<
                /Type /Font
                /Subtype /Type1
                /Name /F1
                /BaseFont /Helvetica
                /Encoding /WinAnsiEncoding
            >>"""))

    width, height = 100, 100
    image_data = b"\xff\x7f\x00" * (width * height)

    image = qpdf.Stream(pdf, image_data)
    image.stream_dict = qpdf.Object.parse(b"""
            <<
                /Type /XObject
                /Subtype /Image
                /ColorSpace /DeviceRGB
                /BitsPerComponent 8
                /Width 100
                /Height 100
            >>""")

    rfont = {'/F1': font}

    xobj = {'/Im1': image}

    resources = {
        '/Font': rfont,
        '/XObject': xobj
        }

    mediabox = [0, 0, 612, 792]

    stream = b"""
        BT /F1 24 Tf 72 720 Td (Hi there) Tj ET
        q 144 0 0 144 234 324 cm /Im1 Do Q
        """

    contents = qpdf.Stream(pdf, stream)

    page_dict = {
        '/Type': qpdf.Name('/Page'),
        '/MediaBox': mediabox,
        '/Contents': contents,
        '/Resources': resources
        }
    qpdf_page_dict = page_dict
    page = pdf.make_indirect(qpdf_page_dict)

    pdf.pages.append(page)
    pdf.save(outdir / 'hi.pdf')


def test_copy_semantics(resources):
    pdf = qpdf.Pdf.open(resources / 'graph.pdf')

    # Ensure that we can name a reference to a child object and view the
    # changes from the parent
    page = pdf.pages[0]
    mediabox = page['/MediaBox']
    assert mediabox[2].decode() != 0
    mediabox[2] = 0
    assert page['/MediaBox'][2] == mediabox[2]


def test_save_stream(resources, outdir):
    from io import BytesIO
    pdf = qpdf.Pdf.open(resources / 'graph.pdf')
    pdf.save(outdir / 'nostream.pdf', static_id=True)

    bio = BytesIO()
    pdf.save(bio, static_id=True)
    bio.seek(0)

    with (outdir / 'nostream.pdf').open('rb') as saved_file:
        saved_file_contents = saved_file.read()
    assert saved_file_contents == bio.read()


def test_copy_page_keepalive(resources, outdir):
    # str for py<3.6
    copy(str(resources / 'sandwich.pdf'), str(outdir / 'sandwich.pdf'))
    src = qpdf.Pdf.open(outdir / 'sandwich.pdf')
    pdf = qpdf.Pdf.open(resources / 'graph.pdf')

    pdf.pages.append(src.pages[0])

    del src
    src = None
    gc.collect()
    (outdir / 'sandwich.pdf').unlink()
    pdf.save(outdir / 'out.pdf')




import pytest
from pikepdf import _qpdf as qpdf
import os
import platform
import shutil
from pathlib import Path
from contextlib import suppress
from subprocess import run, PIPE, STDOUT, DEVNULL
import xml.etree.ElementTree as ET


VERAPDF = Path(os.environ['HOME']) / 'verapdf' / 'verapdf'


NO_PDFA_VALIDATOR = not VERAPDF.is_file()


def verapdf_validate(filename):
    with open(filename, 'rb') as f:
        proc = run([VERAPDF], stdin=f, stdout=PIPE, stderr=STDOUT, check=True)
        result = proc.stdout.decode('utf-8')

        xml_start = result.find('<?xml version')
        xml = result[xml_start:]

    root = ET.fromstring(xml)
    node = root.find(".//taskResult[@type='VALIDATE']")
    return node.attrib['isExecuted'] == 'true' and \
            node.attrib['isSuccess'] == 'true'


@pytest.mark.skipif(NO_PDFA_VALIDATOR, reason="can't find verapdf")
def test_pdfa_sanity(resources, outdir):
    filename = resources / 'veraPDF test suite 6-2-10-t02-pass-a.pdf'

    assert verapdf_validate(filename)

    pdf = qpdf.Pdf.open(filename)
    pdf.save(outdir / 'pdfa.pdf')

    assert verapdf_validate(outdir / 'pdfa.pdf')






# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)

from io import BytesIO
from subprocess import run, PIPE
from ._qpdf import Pdf, Object, ObjectType, Array, Name
from tempfile import NamedTemporaryFile


class DependencyError(Exception):
    pass


class PdfImage:
    def __init__(self, obj):
        if obj.type_code not in (ObjectType.stream, ObjectType.inlineimage):
            raise TypeError("can't construct PdfImage from non-image")

        if obj.type_code == ObjectType.stream and \
                obj.stream_dict.get("/Subtype") != "/Image":
            raise TypeError("can't construct PdfImage from non-image")
        
        self.obj = obj

    @property
    def mode(self):
        m = ''
        if self.obj.BitsPerComponent == 1:
            m = '1'
        elif self.obj.BitsPerComponent == 8:
            if self.obj.ColorSpace == '/DeviceRGB':
                m = 'RGB'
            elif self.obj.ColorSpace == '/DeviceGray':
                m = 'L'
        if m == '':
            raise NotImplementedError("Not sure how to handle PDF image of this type")
        return m

    @property
    def size(self):
        return self.width, self.height

    @property
    def width(self):
        return int(self.obj.Width)

    @property
    def height(self):
        return int(self.obj.Height)

    def topil(self):
        from PIL import Image

        im = None
        if self.obj.Filter == '/DCTDecode' or \
                self.obj.Filter == Array([Name('/DCTDecode')]):
            buffer = self.obj.get_raw_stream_buffer()
            # BytesIO will make a copy of buffer
            im = Image.open(BytesIO(buffer))  
            return im
                
        if self.mode == 'RGB':
            # No point in accessing the buffer here, size qpdf decodes to 3-byte
            # RGB and Pillow needs RGBX for raw access
            data = self.obj.read_bytes()
            im = Image.frombytes('RGB', self.size, data)
        elif self.mode == 'L':
            buffer = self.obj.get_stream_buffer()
            stride = 0  # tell Pillow to calculate stride from line width
            ystep = 1  # image is top to bottom in memory
            im = Image.frombuffer('L', self.size, buffer, "raw", stride, ystep)
        
        if not im:
            raise TypeError("don't how to convert to this image type")

        return im
    
    def show(self):
        self.topil().show()

    def __repr__(self):
        return '<pikepdf.PdfImage image mode={} size={}x{} at {}>'.format(
            self.mode, self.width, self.height, hex(id(self)))

    def _repr_png_(self):
        "Display hook for IPython/Jupyter"
        b = BytesIO()
        im = self.topil()
        im.save(b, 'PNG')
        return b.getvalue()


def page_to_svg(page):
    pdf = Pdf.new()
    pdf.pages.append(page)
    with NamedTemporaryFile(suffix='.pdf') as tmp_in, \
            NamedTemporaryFile(mode='w+b', suffix='.svg') as tmp_out:
        pdf.save(tmp_in)
        tmp_in.seek(0)

        try:
            proc = run(['mudraw', '-F', 'svg', '-o', tmp_out.name, tmp_in.name], stderr=PIPE)
        except FileNotFoundError as e:
            raise DependencyError("Could not find the required executable 'mutool'")

        if proc.stderr:
            print(proc.stderr.decode())            
        tmp_out.flush()
        tmp_out2 = open(tmp_out.name, 'rb')  # Not sure why re-opening is need, but it is
        svg = tmp_out2.read()
        return svg.decode()